# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import asyncio
import inspect
import queue
import threading
from typing import Optional, Callable, Union, TYPE_CHECKING
import logging
from datetime import datetime

import websockets

from ...utils import verboselogs
from .constants import LOGGING, CHANNELS, RATE, CHUNK, TIMEOUT, PLAYBACK_DELTA

from ..microphone import Microphone

if TYPE_CHECKING:
    import pyaudio

HALF_SECOND = 0.5


class Speaker:  # pylint: disable=too-many-instance-attributes
    """
    This implements a speaker for local audio output. This uses PyAudio under the hood.
    """

    _logger: verboselogs.VerboseLogger

    _audio: Optional["pyaudio.PyAudio"] = None
    _stream: Optional["pyaudio.Stream"] = None

    _chunk: int
    _rate: int
    _channels: int
    _output_device_index: Optional[int] = None

    # last time we received audio
    _last_datagram: datetime = datetime.now()
    _last_play_delta_in_ms: int
    _lock_wait: threading.Lock

    _queue: queue.Queue
    _exit: threading.Event

    _thread: Optional[threading.Thread] = None
    # _asyncio_loop: asyncio.AbstractEventLoop
    # _asyncio_thread: threading.Thread
    _receiver_thread: Optional[threading.Thread] = None
    _loop: Optional[asyncio.AbstractEventLoop] = None

    _push_callback_org: Optional[Callable] = None
    _push_callback: Optional[Callable] = None
    _pull_callback_org: Optional[Callable] = None
    _pull_callback: Optional[Callable] = None

    _microphone: Optional[Microphone] = None

    def __init__(
        self,
        pull_callback: Optional[Callable] = None,
        push_callback: Optional[Callable] = None,
        verbose: int = LOGGING,
        rate: int = RATE,
        chunk: int = CHUNK,
        channels: int = CHANNELS,
        last_play_delta_in_ms: int = PLAYBACK_DELTA,
        output_device_index: Optional[int] = None,
        microphone: Optional[Microphone] = None,
    ):  # pylint: disable=too-many-positional-arguments
        # dynamic import of pyaudio as not to force the requirements on the SDK (and users)
        import pyaudio  # pylint: disable=import-outside-toplevel

        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(verbose)

        self._exit = threading.Event()
        self._queue = queue.Queue()

        self._last_datagram = datetime.now()
        self._lock_wait = threading.Lock()

        self._microphone = microphone

        self._audio = pyaudio.PyAudio()
        self._chunk = chunk
        self._rate = rate
        self._format = pyaudio.paInt16
        self._channels = channels
        self._last_play_delta_in_ms = last_play_delta_in_ms
        self._output_device_index = output_device_index

        self._push_callback_org = push_callback
        self._pull_callback_org = pull_callback

    def set_push_callback(self, push_callback: Callable) -> None:
        """
        set_push_callback - sets the callback function to be called when data is sent.

        Args:
            push_callback (Callable): The callback function to be called when data is send.
                                      This should be the websocket handle message function.

        Returns:
            None
        """
        self._push_callback_org = push_callback

    def set_pull_callback(self, pull_callback: Callable) -> None:
        """
        set_pull_callback - sets the callback function to be called when data is received.

        Args:
            pull_callback (Callable): The callback function to be called when data is received.
                                      This should be the websocket recv function.

        Returns:
            None
        """
        self._pull_callback_org = pull_callback

    def start(self, active_loop: Optional[asyncio.AbstractEventLoop] = None) -> bool:
        """
        starts - starts the Speaker stream

        Args:
            socket (Union[SyncClientConnection, AsyncClientConnection]): The socket to receive audio data from.

        Returns:
            bool: True if the stream was started, False otherwise
        """
        self._logger.debug("Speaker.start ENTER")

        self._logger.info("format: %s", self._format)
        self._logger.info("channels: %d", self._channels)
        self._logger.info("rate: %d", self._rate)
        self._logger.info("chunk: %d", self._chunk)
        # self._logger.info("output_device_id: %d", self._output_device_index)

        # Automatically get the current running event loop
        if inspect.iscoroutinefunction(self._push_callback_org) and active_loop is None:
            self._logger.verbose("get default running asyncio loop")
            self._loop = asyncio.get_running_loop()

        self._exit.clear()
        self._queue = queue.Queue()

        if self._audio is not None:
            self._stream = self._audio.open(
                format=self._format,
                channels=self._channels,
                rate=self._rate,
                input=False,
                output=True,
                frames_per_buffer=self._chunk,
                output_device_index=self._output_device_index,
            )

        if self._stream is None:
            self._logger.error("start failed. No stream created.")
            self._logger.debug("Speaker.start LEAVE")
            return False

        self._push_callback = self._push_callback_org
        self._pull_callback = self._pull_callback_org

        # start the play thread
        self._thread = threading.Thread(
            target=self._play, args=(self._queue, self._stream, self._exit), daemon=True
        )
        self._thread.start()

        # Start the stream
        if self._stream is not None:
            self._stream.start_stream()

        # Start the receiver thread within the start function
        self._logger.verbose("Starting receiver thread...")
        self._receiver_thread = threading.Thread(target=self._start_receiver)
        self._receiver_thread.start()

        self._logger.notice("start succeeded")
        self._logger.debug("Speaker.start LEAVE")

        return True

    def wait_for_complete_with_mute(self, mic: Microphone):
        """
        This method will mute/unmute a Microphone and block until the speak is done playing sound.
        """
        self._logger.debug("Speaker.wait_for_complete ENTER")

        if self._microphone is not None:
            mic.mute()
        self.wait_for_complete()
        if self._microphone is not None:
            mic.unmute()

        self._logger.debug("Speaker.wait_for_complete LEAVE")

    def wait_for_complete(self):
        """
        This method will block until the speak is done playing sound.
        """
        self._logger.debug("Speaker.wait_for_complete ENTER")

        delta_in_ms = float(self._last_play_delta_in_ms)
        self._logger.debug("Last Play delta: %f", delta_in_ms)

        # set to now
        with self._lock_wait:
            self._last_datagram = datetime.now()

        while True:
            # sleep for a bit
            self._exit.wait(HALF_SECOND)

            # check if we should exit
            if self._exit.is_set():
                self._logger.debug("Exiting wait_for_complete _exit is set")
                break

            # check the time
            with self._lock_wait:
                delta = datetime.now() - self._last_datagram
                diff_in_ms = delta.total_seconds() * 1000
                if diff_in_ms < delta_in_ms:
                    self._logger.debug("LastPlay delta is less than threshold")
                    continue

            # if we get here, we are done playing audio
            self._logger.debug("LastPlay delta is greater than threshold. Exit wait!")
            break

        self._logger.debug("Speaker.wait_for_complete LEAVE")

    def _start_receiver(self):
        # Check if the socket is an asyncio WebSocket
        if inspect.iscoroutinefunction(self._pull_callback_org):
            self._logger.verbose("Starting asyncio receiver...")
            asyncio.run_coroutine_threadsafe(self._start_asyncio_receiver(), self._loop)
        else:
            self._logger.verbose("Starting threaded receiver...")
            self._start_threaded_receiver()

    async def _start_asyncio_receiver(self):
        try:
            while True:
                if self._exit.is_set():
                    self._logger.verbose("Exiting receiver thread...")
                    break

                message = await self._pull_callback()
                if message is None:
                    self._logger.verbose("No message received...")
                    continue

                if isinstance(message, str):
                    self._logger.verbose("Received control message...")
                    await self._push_callback(message)
                elif isinstance(message, bytes):
                    self._logger.verbose("Received audio data...")
                    await self._push_callback(message)
                    self.add_audio_to_queue(message)
        except websockets.exceptions.ConnectionClosedOK as e:
            self._logger.debug("send() exiting gracefully: %d", e.code)
        except websockets.exceptions.ConnectionClosed as e:
            if e.code in [1000, 1001]:
                self._logger.debug("send() exiting gracefully: %d", e.code)
                return
            self._logger.error("_start_asyncio_receiver - ConnectionClosed: %s", str(e))
        except websockets.exceptions.WebSocketException as e:
            self._logger.error(
                "_start_asyncio_receiver- WebSocketException: %s", str(e)
            )
        except Exception as e:  # pylint: disable=broad-except
            self._logger.error("_start_asyncio_receiver exception: %s", str(e))

    def _start_threaded_receiver(self):
        try:
            while True:
                if self._exit.is_set():
                    self._logger.verbose("Exiting receiver thread...")
                    break

                message = self._pull_callback()
                if message is None:
                    self._logger.verbose("No message received...")
                    continue

                if isinstance(message, str):
                    self._logger.verbose("Received control message...")
                    self._push_callback(message)
                elif isinstance(message, bytes):
                    self._logger.verbose("Received audio data...")
                    self._push_callback(message)
                    self.add_audio_to_queue(message)
        except Exception as e:  # pylint: disable=broad-except
            self._logger.notice("_start_threaded_receiver exception: %s", str(e))

    def add_audio_to_queue(self, data: bytes) -> None:
        """
        add_audio_to_queue - adds audio data to the Speaker queue

        Args:
            data (bytes): The audio data to add to the queue
        """
        self._queue.put(data)

    def finish(self) -> bool:
        """
        finish - stops the Speaker stream

        Returns:
            bool: True if the stream was stopped, False otherwise
        """
        self._logger.debug("Speaker.finish ENTER")

        self._logger.notice("signal exit")
        self._exit.set()

        if self._stream is not None:
            self._logger.notice("stopping stream...")
            self._stream.stop_stream()
            self._stream.close()
            self._logger.notice("stream stopped")

        if self._thread is not None:
            self._logger.notice("joining _thread...")
            self._thread.join()
            self._logger.notice("thread stopped")

        if self._receiver_thread is not None:
            self._logger.notice("stopping _receiver_thread...")
            self._receiver_thread.join()
            self._logger.notice("_receiver_thread joined")

        with self._queue.mutex:
            self._queue.queue.clear()

        self._stream = None
        self._thread = None
        self._receiver_thread = None

        self._logger.notice("finish succeeded")
        self._logger.debug("Speaker.finish LEAVE")

        return True

    def _play(self, audio_out, stream, stop):
        """
        _play - plays audio data from the Speaker queue callback for portaudio
        """
        while not stop.is_set():
            try:
                if self._microphone is not None and self._microphone.is_muted():
                    with self._lock_wait:
                        delta = datetime.now() - self._last_datagram
                        diff_in_ms = delta.total_seconds() * 1000
                        if diff_in_ms > float(self._last_play_delta_in_ms):
                            self._logger.debug(
                                "LastPlay delta is greater than threshold. Unmute!"
                            )
                            self._microphone.unmute()

                data = audio_out.get(True, TIMEOUT)
                with self._lock_wait:
                    self._last_datagram = datetime.now()
                    if self._microphone is not None and not self._microphone.is_muted():
                        self._logger.debug("New speaker sound detected. Mute!")
                        self._microphone.mute()
                stream.write(data)
            except queue.Empty:
                pass
            except Exception as e:  # pylint: disable=broad-except
                self._logger.error("_play exception: %s", str(e))



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from typing import List, Optional, Dict, Any

from dataclasses import dataclass

# common websocket response
from ....common import (
    BaseResponse,
    OpenResponse,
    CloseResponse,
    ErrorResponse,
    UnhandledResponse,
)

# unique


@dataclass
class WelcomeResponse(BaseResponse):
    """
    The server will send a Welcome message as soon as the websocket opens.
    """

    type: str
    request_id: str


@dataclass
class SettingsAppliedResponse(BaseResponse):
    """
    The server will send a SettingsApplied message as soon as the settings are applied.
    """

    type: str


@dataclass
class ConversationTextResponse(BaseResponse):
    """
    The server will send a ConversationText message every time the agent hears the user say something, and every time the agent speaks something itself.
    """

    type: str
    role: str
    content: str


@dataclass
class UserStartedSpeakingResponse(BaseResponse):
    """
    The server will send a UserStartedSpeaking message every time the user begins a new utterance.
    """

    type: str


@dataclass
class AgentThinkingResponse(BaseResponse):
    """
    The server will send an AgentThinking message to inform the client of a non-verbalized agent thought.
    You will ONLY receive this message if you have set `experimental` to true.
    """

    type: str
    content: str


@dataclass
class FunctionCallRequest(BaseResponse):
    """
    The FunctionCallRequest message is used to call a function from the server to the client.
    """

    type: str
    function_name: str
    function_call_id: str
    input: str


@dataclass
class AgentStartedSpeakingResponse(BaseResponse):
    """
    The server will send an AgentStartedSpeaking message when it begins streaming an agent audio response to the client for playback.
    """

    total_latency: float
    tts_latency: float
    ttt_latency: float


@dataclass
class AgentAudioDoneResponse(BaseResponse):
    """
    The server will send an AgentAudioDone message immediately after it sends the last audio message in a piece of agent speech.
    """

    type: str


@dataclass
class InjectionRefusedResponse(BaseResponse):
    """
    The server will send an InjectionRefused message when an InjectAgentMessage request is ignored because it arrived while the user was speaking or while the server was sending audio for an agent response.
    """

    type: str



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .enums import LiveTranscriptionEvents

# backward compat
from .client import (
    PreRecordedClient,
    AsyncPreRecordedClient,
    LiveClient,
    AsyncLiveClient,
)

# rest
# common
from .client import (
    UrlSource,
    TextSource,
    BufferSource,
    StreamSource,
    FileSource,
)

## input
from .client import (
    ListenRESTOptions,
    PrerecordedOptions,
    PreRecordedStreamSource,
    PrerecordedSource,
    ListenRestSource,
)

## output
from .client import (
    # top level
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
    # shared
    Average,
    Intent,
    Intents,
    IntentsInfo,
    Segment,
    SentimentInfo,
    Sentiment,
    Sentiments,
    SummaryInfo,
    Topic,
    Topics,
    TopicsInfo,
    # between rest and websocket
    ModelInfo,
    Hit,
    Search,
    # unique
    Entity,
    ListenRESTMetadata,
    Paragraph,
    Paragraphs,
    ListenRESTResults,
    Sentence,
    Summaries,
    SummaryV1,
    SummaryV2,
    Translation,
    Utterance,
    Warning,
    ListenRESTAlternative,
    ListenRESTChannel,
    ListenRESTWord,
)


# websocket
## input
from .client import (
    ListenWebSocketOptions,
    LiveOptions,
)

## output
from .client import (
    # top level
    LiveResultResponse,
    ListenWSMetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    # common websocket response
    OpenResponse,
    CloseResponse,
    ErrorResponse,
    UnhandledResponse,
    # unique
    ListenWSMetadata,
    ListenWSWord,
    ListenWSAlternative,
    ListenWSChannel,
)

# clients
from .client import (
    ListenRESTClient,
    AsyncListenRESTClient,
    ListenWebSocketClient,
    AsyncListenWebSocketClient,
)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import json
import time
import logging
from typing import Dict, Union, Optional, cast, Any, Callable
from datetime import datetime
import threading

from .....utils import verboselogs
from .....options import DeepgramClientOptions
from ...enums import SpeakWebSocketEvents, SpeakWebSocketMessage
from ....common import AbstractSyncWebSocketClient
from ....common import DeepgramError

from .response import (
    OpenResponse,
    MetadataResponse,
    FlushedResponse,
    ClearedResponse,
    CloseResponse,
    WarningResponse,
    ErrorResponse,
    UnhandledResponse,
)
from .options import SpeakWSOptions

from .....audio.microphone import Microphone
from .....audio.speaker import Speaker, RATE, CHANNELS, PLAYBACK_DELTA

ONE_SECOND = 1
HALF_SECOND = 0.5
DEEPGRAM_INTERVAL = 5
PING_INTERVAL = 20


class SpeakWSClient(
    AbstractSyncWebSocketClient
):  # pylint: disable=too-many-instance-attributes
    """
    Client for interacting with Deepgram's text-to-speech services over WebSockets.

     This class provides methods to establish a WebSocket connection for TTS synthesis and handle real-time TTS synthesis events.

     Args:
         config (DeepgramClientOptions): all the options for the client.
    """

    _logger: verboselogs.VerboseLogger
    _config: DeepgramClientOptions
    _endpoint: str

    _event_handlers: Dict[SpeakWebSocketEvents, list]

    _flush_thread: Union[threading.Thread, None]
    _lock_flush: threading.Lock
    _last_datagram: Optional[datetime] = None
    _flush_count: int

    _kwargs: Optional[Dict] = None
    _addons: Optional[Dict] = None
    _options: Optional[Dict] = None
    _headers: Optional[Dict] = None

    _speaker_created: bool = False
    _speaker: Optional[Speaker] = None
    _microphone: Optional[Microphone] = None

    def __init__(
        self, config: DeepgramClientOptions, microphone: Optional[Microphone] = None
    ):
        if config is None:
            raise DeepgramError("Config is required")

        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(config.verbose)

        self._config = config
        self._endpoint = "v1/speak"
        self._lock_flush = threading.Lock()

        self._flush_thread = None

        # auto flush
        self._last_datagram = None
        self._flush_count = 0

        # microphone
        self._microphone = microphone

        # init handlers
        self._event_handlers = {
            event: [] for event in SpeakWebSocketEvents.__members__.values()
        }

        if self._config.options.get("speaker_playback") == "true":
            self._logger.info("speaker_playback is enabled")
            rate = self._config.options.get("speaker_playback_rate")
            if rate is None:
                rate = RATE
            channels = self._config.options.get("speaker_playback_channels")
            if channels is None:
                channels = CHANNELS
            playback_delta_in_ms = self._config.options.get(
                "speaker_playback_delta_in_ms"
            )
            if playback_delta_in_ms is None:
                playback_delta_in_ms = PLAYBACK_DELTA
            device_index = self._config.options.get("speaker_playback_device_index")

            self._logger.debug("rate: %s", rate)
            self._logger.debug("channels: %s", channels)
            self._logger.debug("device_index: %s", device_index)

            self._speaker_created = True

            if device_index is not None:
                self._speaker = Speaker(
                    rate=rate,
                    channels=channels,
                    last_play_delta_in_ms=playback_delta_in_ms,
                    verbose=self._config.verbose,
                    output_device_index=device_index,
                    microphone=self._microphone,
                )
            else:
                self._speaker = Speaker(
                    rate=rate,
                    channels=channels,
                    last_play_delta_in_ms=playback_delta_in_ms,
                    verbose=self._config.verbose,
                    microphone=self._microphone,
                )

        # call the parent constructor
        super().__init__(self._config, self._endpoint)

    # pylint: disable=too-many-statements,too-many-branches
    def start(
        self,
        options: Optional[Union[SpeakWSOptions, Dict]] = None,
        addons: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        members: Optional[Dict] = None,
        **kwargs,
    ) -> bool:
        """
        Starts the WebSocket connection for text-to-speech synthesis.
        """
        self._logger.debug("SpeakWebSocketClient.start ENTER")
        self._logger.info("options: %s", options)
        self._logger.info("addons: %s", addons)
        self._logger.info("headers: %s", headers)
        self._logger.info("members: %s", members)
        self._logger.info("kwargs: %s", kwargs)

        if isinstance(options, SpeakWSOptions) and not options.check():
            self._logger.error("options.check failed")
            self._logger.debug("SpeakWebSocketClient.start LEAVE")
            raise DeepgramError("Fatal text-to-speech options error")

        self._addons = addons
        self._headers = headers

        # add "members" as members of the class
        if members is not None:
            self.__dict__.update(members)

        # set kwargs as members of the class
        if kwargs is not None:
            self._kwargs = kwargs
        else:
            self._kwargs = {}

        if isinstance(options, SpeakWSOptions):
            self._logger.info("SpeakWSOptions switching class -> dict")
            self._options = options.to_dict()
        elif options is not None:
            self._options = options
        else:
            self._options = {}

        try:
            # speaker substitutes the listening thread
            if self._speaker is not None:
                self._logger.notice("passing speaker to delegate_listening")
                super().delegate_listening(self._speaker)

            # call parent start
            if (
                super().start(
                    self._options,
                    self._addons,
                    self._headers,
                    **dict(cast(Dict[Any, Any], self._kwargs)),
                )
                is False
            ):
                self._logger.error("SpeakWebSocketClient.start failed")
                self._logger.debug("SpeakWebSocketClient.start LEAVE")
                return False

            if self._speaker is not None:
                self._logger.notice("start delegate_listening thread")
                self._speaker.start()

            # debug the threads
            for thread in threading.enumerate():
                self._logger.debug("after running thread: %s", thread.name)
            self._logger.debug("number of active threads: %s", threading.active_count())

            # flush thread
            if self._config.is_auto_flush_speak_enabled():
                self._logger.notice("autoflush is enabled")
                self._flush_thread = threading.Thread(target=self._flush)
                self._flush_thread.start()
            else:
                self._logger.notice("autoflush is disabled")

            # debug the threads
            for thread in threading.enumerate():
                self._logger.debug("after running thread: %s", thread.name)
            self._logger.debug("number of active threads: %s", threading.active_count())

            self._logger.notice("start succeeded")
            self._logger.debug("SpeakWebSocketClient.start LEAVE")
            return True

        except Exception as e:  # pylint: disable=broad-except
            self._logger.error(
                "WebSocketException in SpeakWebSocketClient.start: %s", e
            )
            self._logger.debug("SpeakWebSocketClient.start LEAVE")
            if self._config.options.get("termination_exception_connect") is True:
                raise
            return False

    # pylint: enable=too-many-statements,too-many-branches

    def on(self, event: SpeakWebSocketEvents, handler: Callable) -> None:
        """
        Registers event handlers for specific events.
        """
        self._logger.info("event subscribed: %s", event)
        if event in SpeakWebSocketEvents.__members__.values() and callable(handler):
            self._event_handlers[event].append(handler)

    def _emit(self, event: SpeakWebSocketEvents, *args, **kwargs) -> None:
        """
        Emits events to the registered event handlers.
        """
        self._logger.debug("SpeakWebSocketClient._emit ENTER")
        self._logger.debug("callback handlers for: %s", event)

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("after running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        self._logger.debug("callback handlers for: %s", event)
        for handler in self._event_handlers[event]:
            handler(self, *args, **kwargs)

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("after running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        self._logger.debug("ListenWebSocketClient._emit LEAVE")

    def _process_text(self, message: str) -> None:
        """
        Processes messages received over the WebSocket connection.
        """
        self._logger.debug("SpeakWebSocketClient._process_text ENTER")

        try:
            self._logger.debug("Text data received")

            if len(message) == 0:
                self._logger.debug("message is empty")
                self._logger.debug("SpeakWebSocketClient._process_text LEAVE")
                return

            data = json.loads(message)
            response_type = data.get("type")
            self._logger.debug("response_type: %s, data: %s", response_type, data)

            match response_type:
                case SpeakWebSocketEvents.Open:
                    open_result: OpenResponse = OpenResponse.from_json(message)
                    self._logger.verbose("OpenResponse: %s", open_result)
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Open),
                        open=open_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case SpeakWebSocketEvents.Metadata:
                    meta_result: MetadataResponse = MetadataResponse.from_json(message)
                    self._logger.verbose("MetadataResponse: %s", meta_result)
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Metadata),
                        metadata=meta_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case SpeakWebSocketEvents.Flushed:
                    fl_result: FlushedResponse = FlushedResponse.from_json(message)
                    self._logger.verbose("FlushedResponse: %s", fl_result)

                    # auto flush
                    if self._config.is_inspecting_speak():
                        with self._lock_flush:
                            self._flush_count -= 1
                            self._logger.debug(
                                "Decrement Flush count: %d",
                                self._flush_count,
                            )

                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Flushed),
                        flushed=fl_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case SpeakWebSocketEvents.Cleared:
                    clear_result: ClearedResponse = ClearedResponse.from_json(message)
                    self._logger.verbose("ClearedResponse: %s", clear_result)
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Cleared),
                        cleared=clear_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case SpeakWebSocketEvents.Close:
                    close_result: CloseResponse = CloseResponse.from_json(message)
                    self._logger.verbose("CloseResponse: %s", close_result)
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Close),
                        close=close_result,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case SpeakWebSocketEvents.Warning:
                    war_warning: WarningResponse = WarningResponse.from_json(message)
                    self._logger.verbose("WarningResponse: %s", war_warning)
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Warning),
                        warning=war_warning,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case SpeakWebSocketEvents.Error:
                    err_error: ErrorResponse = ErrorResponse.from_json(message)
                    self._logger.verbose("ErrorResponse: %s", err_error)
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Error),
                        error=err_error,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )
                case _:
                    self._logger.warning(
                        "Unknown Message: response_type: %s, data: %s",
                        response_type,
                        data,
                    )
                    unhandled_error: UnhandledResponse = UnhandledResponse(
                        type=SpeakWebSocketEvents(SpeakWebSocketEvents.Unhandled),
                        raw=message,
                    )
                    self._emit(
                        SpeakWebSocketEvents(SpeakWebSocketEvents.Unhandled),
                        unhandled=unhandled_error,
                        **dict(cast(Dict[Any, Any], self._kwargs)),
                    )

            self._logger.notice("_process_text Succeeded")
            self._logger.debug("SpeakWebSocketClient._process_text LEAVE")

        except Exception as e:  # pylint: disable=broad-except
            self._logger.error("Exception in SpeakWebSocketClient._process_text: %s", e)
            e_error: ErrorResponse = ErrorResponse(
                "Exception in SpeakWebSocketClient._process_text",
                f"{e}",
                "Exception",
            )
            self._logger.error(
                "Exception in SpeakWebSocketClient._process_text: %s", str(e)
            )
            self._emit(
                SpeakWebSocketEvents(SpeakWebSocketEvents.Error),
                e_error,
                **dict(cast(Dict[Any, Any], self._kwargs)),
            )

            # signal exit and close
            super()._signal_exit()

            self._logger.debug("SpeakWebSocketClient._process_text LEAVE")

            if self._config.options.get("termination_exception") is True:
                raise
            return

    # pylint: enable=too-many-return-statements,too-many-statements

    def _process_binary(self, message: bytes) -> None:
        self._logger.debug("SpeakWebSocketClient._process_binary ENTER")
        self._logger.debug("Binary data received")

        self._emit(
            SpeakWebSocketEvents(SpeakWebSocketEvents.AudioData),
            data=message,
            **dict(cast(Dict[Any, Any], self._kwargs)),
        )

        self._logger.notice("_process_binary Succeeded")
        self._logger.debug("SpeakWebSocketClient._process_binary LEAVE")

    # pylint: disable=too-many-return-statements
    def _flush(self) -> None:
        self._logger.debug("SpeakWebSocketClient._flush ENTER")

        delta_in_ms_str = self._config.options.get("auto_flush_speak_delta")
        if delta_in_ms_str is None:
            self._logger.error("auto_flush_speak_delta is None")
            self._logger.debug("SpeakWebSocketClient._flush LEAVE")
            return
        delta_in_ms = float(delta_in_ms_str)

        while True:
            try:
                self._exit_event.wait(timeout=HALF_SECOND)

                if self._exit_event.is_set():
                    self._logger.notice("_flush exiting gracefully")
                    self._logger.debug("ListenWebSocketClient._flush LEAVE")
                    return

                if self._last_datagram is None:
                    self._logger.debug("AutoFlush last_datagram is None")
                    continue

                with self._lock_flush:
                    delta = datetime.now() - self._last_datagram
                    diff_in_ms = delta.total_seconds() * 1000
                    self._logger.debug("AutoFlush delta: %f", diff_in_ms)
                    if diff_in_ms < delta_in_ms:
                        self._logger.debug("AutoFlush delta is less than threshold")
                        continue

                self.flush()

            except Exception as e:  # pylint: disable=broad-except
                self._logger.error("Exception in SpeakWebSocketClient._flush: %s", e)
                e_error: ErrorResponse = ErrorResponse(
                    "Exception in SpeakWebSocketClient._flush",
                    f"{e}",
                    "Exception",
                )
                self._logger.error(
                    "Exception in SpeakWebSocketClient._flush: %s", str(e)
                )
                self._emit(
                    SpeakWebSocketEvents(SpeakWebSocketEvents.Error),
                    error=e_error,
                    **dict(cast(Dict[Any, Any], self._kwargs)),
                )

                # signal exit and close
                super()._signal_exit()

                self._logger.debug("SpeakWebSocketClient._flush LEAVE")

                if self._config.options.get("termination_exception") is True:
                    raise
                return

    # pylint: enable=too-many-return-statements

    def send_text(self, text_input: str) -> bool:
        """
        Sends text to the WebSocket connection to generate audio.

        Args:
            text_input (str): The raw text to be synthesized. This function will automatically wrap
                the text in a JSON object of type "Speak" with the key "text".

        Returns:
            bool: True if the text was successfully sent, False otherwise.
        """
        return self.send_raw(json.dumps({"type": "Speak", "text": text_input}))

    def send(self, data: Union[str, bytes]) -> bool:
        """
        Alias for send_text. Please see send_text for more information.
        """
        if isinstance(data, bytes):
            self._logger.error("send() failed - data is bytes")
            return False

        return self.send_text(data)

    # pylint: disable=unused-argument
    def send_control(
        self, msg_type: Union[SpeakWebSocketMessage, str], data: Optional[str] = ""
    ) -> bool:
        """
        Sends a control message consisting of type SpeakWebSocketEvents over the WebSocket connection.

        Args:
            msg_type (SpeakWebSocketEvents): The type of control message to send.
            (Optional) data (str): The data to send with the control message.

        Returns:
            bool: True if the control message was successfully sent, False otherwise.
        """
        control_msg = json.dumps({"type": msg_type})
        return self.send_raw(control_msg)

    # pylint: enable=unused-argument

    # pylint: disable=too-many-return-statements,too-many-branches,too-many-statements
    def send_raw(self, msg: str) -> bool:
        """
        Sends a raw/control message over the WebSocket connection. This message must contain a valid JSON object.

        Args:
            msg (str): The raw message to send over the WebSocket connection.

        Returns:
            bool: True if the message was successfully sent, False otherwise.
        """
        self._logger.spam("SpeakWebSocketClient.send_raw ENTER")

        if self._config.is_inspecting_speak():
            try:
                _tmp_json = json.loads(msg)
                if "type" in _tmp_json:
                    self._logger.debug(
                        "Inspecting Message: Sending %s", _tmp_json["type"]
                    )
                    match _tmp_json["type"]:
                        case SpeakWebSocketMessage.Speak:
                            inspect_res = self._inspect()
                            if not inspect_res:
                                self._logger.error("inspect_res failed")
                        case SpeakWebSocketMessage.Flush:
                            with self._lock_flush:
                                self._last_datagram = None
                                self._flush_count += 1
                                self._logger.debug(
                                    "Increment Flush count: %d", self._flush_count
                                )
            except Exception as e:  # pylint: disable=broad-except
                self._logger.error("send_raw() failed - Exception: %s", str(e))

        try:
            if super().send(msg) is False:
                self._logger.error("send_raw() failed")
                self._logger.spam("SpeakWebSocketClient.send_raw LEAVE")
                return False
            self._logger.spam("send_raw() succeeded")
            self._logger.spam("SpeakWebSocketClient.send_raw LEAVE")
            return True
        except Exception as e:  # pylint: disable=broad-except
            self._logger.error("send_raw() failed - Exception: %s", str(e))
            self._logger.spam("SpeakWebSocketClient.send_raw LEAVE")
            if self._config.options.get("termination_exception_send") is True:
                raise
            return False

    # pylint: enable=too-many-return-statements,too-many-branches

    def flush(self) -> bool:
        """
        Flushes the current buffer and returns generated audio
        """
        self._logger.spam("SpeakWebSocketClient.flush ENTER")

        self._logger.notice("Sending Flush...")
        ret = self.send_control(SpeakWebSocketMessage.Flush)

        if not ret:
            self._logger.error("flush failed")
            self._logger.spam("SpeakWebSocketClient.flush LEAVE")
            return False

        self._logger.notice("flush succeeded")
        self._logger.spam("SpeakWebSocketClient.flush LEAVE")

        return True

    def clear(self) -> bool:
        """
        Clears the current buffer on the server
        """
        self._logger.spam("SpeakWebSocketClient.clear ENTER")

        self._logger.notice("Sending Clear...")
        ret = self.send_control(SpeakWebSocketMessage.Clear)

        if not ret:
            self._logger.error("clear failed")
            self._logger.spam("SpeakWebSocketClient.clear LEAVE")
            return False

        self._logger.notice("clear succeeded")
        self._logger.spam("SpeakWebSocketClient.clear LEAVE")

        return True

    def wait_for_complete(self):
        """
        This method will block until the speak is done playing sound.
        """
        self._logger.spam("SpeakWebSocketClient.wait_for_complete ENTER")

        if self._speaker is None:
            self._logger.error("speaker is None. Return immediately")
            raise DeepgramError("Speaker is not initialized")

        self._speaker.wait_for_complete()
        self._logger.notice("wait_for_complete succeeded")
        self._logger.spam("SpeakWebSocketClient.wait_for_complete LEAVE")

    def _close_message(self) -> bool:
        return self.send_control(SpeakWebSocketMessage.Close)

    # closes the WebSocket connection gracefully
    def finish(self) -> bool:
        """
        Closes the WebSocket connection gracefully.
        """
        self._logger.spam("SpeakWebSocketClient.finish ENTER")

        # call parent finish which calls signal_exit
        if super().finish() is False:
            self._logger.error("ListenWebSocketClient.finish failed")

        if self._speaker is not None and self._speaker_created:
            self._speaker.finish()
            self._speaker_created = False

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("before running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        # stop the threads
        if self._speaker is not None:
            self._logger.verbose("stopping speaker...")
            self._speaker.finish()
            self._speaker = None
            self._logger.notice("speaker stopped")

        if self._flush_thread is not None:
            self._logger.verbose("sdtopping _flush_thread...")
            self._flush_thread.join()
            self._flush_thread = None
            self._logger.notice("_flush_thread joined")

        # debug the threads
        for thread in threading.enumerate():
            self._logger.debug("before running thread: %s", thread.name)
        self._logger.debug("number of active threads: %s", threading.active_count())

        self._logger.notice("finish succeeded")
        self._logger.spam("SpeakWebSocketClient.finish LEAVE")
        return True

    def _inspect(self) -> bool:
        # auto flush_inspect is generically used to track any messages you might want to snoop on
        # place additional logic here to inspect messages of interest

        # for auto flush functionality
        # set the last datagram
        with self._lock_flush:
            self._last_datagram = datetime.now()
            self._logger.debug(
                "AutoFlush last received: %s",
                str(self._last_datagram),
            )

        return True


SpeakWebSocketClient = SpeakWSClient



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import inspect
import asyncio
import threading
from typing import Optional, Callable, Union, TYPE_CHECKING
import logging

from ...utils import verboselogs

from .constants import LOGGING, CHANNELS, RATE, CHUNK

if TYPE_CHECKING:
    import pyaudio


class Microphone:  # pylint: disable=too-many-instance-attributes
    """
    This implements a microphone for local audio input. This uses PyAudio under the hood.
    """

    _logger: verboselogs.VerboseLogger

    _audio: Optional["pyaudio.PyAudio"] = None
    _stream: Optional["pyaudio.Stream"] = None

    _chunk: int
    _rate: int
    _format: int
    _channels: int
    _input_device_index: Optional[int]
    _is_muted: bool

    _asyncio_loop: asyncio.AbstractEventLoop
    _asyncio_thread: Optional[threading.Thread] = None
    _exit: threading.Event

    _push_callback_org: Optional[Callable] = None
    _push_callback: Optional[Callable] = None

    def __init__(
        self,
        push_callback: Optional[Callable] = None,
        verbose: int = LOGGING,
        rate: int = RATE,
        chunk: int = CHUNK,
        channels: int = CHANNELS,
        input_device_index: Optional[int] = None,
    ):  # pylint: disable=too-many-positional-arguments
        # dynamic import of pyaudio as not to force the requirements on the SDK (and users)
        import pyaudio  # pylint: disable=import-outside-toplevel

        self._logger = verboselogs.VerboseLogger(__name__)
        self._logger.addHandler(logging.StreamHandler())
        self._logger.setLevel(verbose)

        self._exit = threading.Event()

        self._audio = pyaudio.PyAudio()
        self._chunk = chunk
        self._rate = rate
        self._format = pyaudio.paInt16
        self._channels = channels
        self._is_muted = False

        self._input_device_index = input_device_index
        self._push_callback_org = push_callback

    def _start_asyncio_loop(self) -> None:
        self._asyncio_loop = asyncio.new_event_loop()
        self._asyncio_loop.run_forever()

    def is_active(self) -> bool:
        """
        is_active - returns the state of the stream

        Args:
            None

        Returns:
            True if the stream is active, False otherwise
        """
        self._logger.debug("Microphone.is_active ENTER")

        if self._stream is None:
            self._logger.error("stream is None")
            self._logger.debug("Microphone.is_active LEAVE")
            return False

        val = self._stream.is_active()
        self._logger.info("is_active: %s", val)
        self._logger.info("is_exiting: %s", self._exit.is_set())
        self._logger.debug("Microphone.is_active LEAVE")
        return val

    def set_callback(self, push_callback: Callable) -> None:
        """
        set_callback - sets the callback function to be called when data is received.

        Args:
            push_callback (Callable): The callback function to be called when data is received.
                                      This should be the websocket send function.

        Returns:
            None
        """
        self._push_callback_org = push_callback

    def start(self) -> bool:
        """
        starts - starts the microphone stream

        Returns:
            bool: True if the stream was started, False otherwise
        """
        self._logger.debug("Microphone.start ENTER")

        self._logger.info("format: %s", self._format)
        self._logger.info("channels: %d", self._channels)
        self._logger.info("rate: %d", self._rate)
        self._logger.info("chunk: %d", self._chunk)
        # self._logger.info("input_device_id: %d", self._input_device_index)

        if self._push_callback_org is None:
            self._logger.error("start failed. No callback set.")
            self._logger.debug("Microphone.start LEAVE")
            return False

        if inspect.iscoroutinefunction(self._push_callback_org):
            self._logger.verbose("async/await callback - wrapping")
            # Run our own asyncio loop.
            self._asyncio_thread = threading.Thread(target=self._start_asyncio_loop)
            self._asyncio_thread.start()

            self._push_callback = lambda data: (
                asyncio.run_coroutine_threadsafe(
                    self._push_callback_org(data), self._asyncio_loop
                ).result()
                if self._push_callback_org
                else None
            )
        else:
            self._logger.verbose("regular threaded callback")
            self._asyncio_thread = None
            self._push_callback = self._push_callback_org

        if self._audio is not None:
            self._stream = self._audio.open(
                format=self._format,
                channels=self._channels,
                rate=self._rate,
                input=True,
                output=False,
                frames_per_buffer=self._chunk,
                input_device_index=self._input_device_index,
                stream_callback=self._callback,
            )

        if self._stream is None:
            self._logger.error("start failed. No stream created.")
            self._logger.debug("Microphone.start LEAVE")
            return False

        self._exit.clear()
        if self._stream is not None:
            self._stream.start_stream()

        self._logger.notice("start succeeded")
        self._logger.debug("Microphone.start LEAVE")
        return True

    def mute(self) -> bool:
        """
        mute - mutes the microphone stream

        Returns:
            bool: True if the stream was muted, False otherwise
        """
        self._logger.verbose("Microphone.mute ENTER")

        if self._stream is None:
            self._logger.error("mute failed. Library not initialized.")
            self._logger.verbose("Microphone.mute LEAVE")
            return False

        self._is_muted = True

        self._logger.notice("mute succeeded")
        self._logger.verbose("Microphone.mute LEAVE")
        return True

    def unmute(self) -> bool:
        """
        unmute - unmutes the microphone stream

        Returns:
            bool: True if the stream was unmuted, False otherwise
        """
        self._logger.verbose("Microphone.unmute ENTER")

        if self._stream is None:
            self._logger.error("unmute failed. Library not initialized.")
            self._logger.verbose("Microphone.unmute LEAVE")
            return False

        self._is_muted = False

        self._logger.notice("unmute succeeded")
        self._logger.verbose("Microphone.unmute LEAVE")
        return True

    def is_muted(self) -> bool:
        """
        is_muted - returns the state of the stream

        Args:
            None

        Returns:
            True if the stream is muted, False otherwise
        """
        self._logger.spam("Microphone.is_muted ENTER")

        if self._stream is None:
            self._logger.spam("is_muted: stream is None")
            self._logger.spam("Microphone.is_muted LEAVE")
            return False

        val = self._is_muted

        self._logger.spam("is_muted: %s", val)
        self._logger.spam("Microphone.is_muted LEAVE")
        return val

    def finish(self) -> bool:
        """
        finish - stops the microphone stream

        Returns:
            bool: True if the stream was stopped, False otherwise
        """
        self._logger.debug("Microphone.finish ENTER")

        self._logger.notice("signal exit")
        self._exit.set()

        # Stop the stream.
        if self._stream is not None:
            self._logger.notice("stopping stream...")
            self._stream.stop_stream()
            self._stream.close()
            self._logger.notice("stream stopped")

        # clean up the thread
        if (
            # inspect.iscoroutinefunction(self._push_callback_org)
            # and
            self._asyncio_thread
            is not None
        ):
            self._logger.notice("stopping _asyncio_loop...")
            self._asyncio_loop.call_soon_threadsafe(self._asyncio_loop.stop)
            self._asyncio_thread.join()
            self._logger.notice("_asyncio_thread joined")
        self._stream = None
        self._asyncio_thread = None

        self._logger.notice("finish succeeded")
        self._logger.debug("Microphone.finish LEAVE")

        return True

    def _callback(
        self, input_data, frame_count, time_info, status_flags
    ):  # pylint: disable=unused-argument
        """
        The callback used to process data in callback mode.
        """
        # dynamic import of pyaudio as not to force the requirements on the SDK (and users)
        import pyaudio  # pylint: disable=import-outside-toplevel

        if self._exit.is_set():
            self._logger.notice("_callback exit is Set. stopping...")
            return None, pyaudio.paAbort

        if input_data is None:
            self._logger.warning("input_data is None")
            return None, pyaudio.paContinue

        try:
            if self._is_muted:
                size = len(input_data)
                input_data = b"\x00" * size

            self._push_callback(input_data)
        except Exception as e:
            self._logger.error("Error while sending: %s", str(e))
            raise

        return input_data, pyaudio.paContinue
